/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/transaction.h"
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransac...
#include "caosdb/entity/v1alpha1/main.pb.h"      // for TransactionRe...
#include "caosdb/file_transmission/download_request_handler.h" // Download...
#include "caosdb/file_transmission/file_reader.h"              // for path
#include "caosdb/file_transmission/register_file_upload_handler.h" // for RegisterFileUploadHandler
#include "caosdb/file_transmission/upload_request_handler.h" // Upload...
#include "caosdb/logging.h"             // for CAOSDB_LOG_FATAL
#include "caosdb/status_code.h"         // for StatusCode
#include "caosdb/transaction_handler.h" // for EntityTransactionHandler
#include <algorithm>                    // for max
#include <boost/filesystem/path.hpp>    // for operator<<, path
#include <boost/log/core/record.hpp>    // for record
#include <boost/log/detail/attachable_sstream_buf.hpp> // for basic_ostring...
#include <boost/log/sources/record_ostream.hpp>        // for basic_record_...
#include <boost/preprocessor/seq/limits/enum_256.hpp>  // for BOOST_PP_SEQ_...
#include <boost/preprocessor/seq/limits/size_256.hpp>  // for BOOST_PP_SEQ_...
// IWYU pragma: no_include <bits/exception.h>
#include <exception>                                // IWYU pragma: keep
#include <google/protobuf/arena.h>                  // for Arena
#include <google/protobuf/generated_message_util.h> // for CreateMessage...
#include <grpc/impl/codegen/gpr_types.h>            // for gpr_timespec
#include <grpcpp/impl/codegen/completion_queue.h>   // for CompletionQueue
#include <iosfwd>                                   // for streamsize
#include <map>                                      // for map, operator!=
#include <memory>                                   // for unique_ptr
#include <stdexcept>                                // for out_of_range
#include <utility>                                  // for move, pair

namespace caosdb {

// TODO(tf) move to another source file.
auto get_status_description(int code) -> const std::string & {
  static const std::string MISSING_DESCRIPTION = "MISSING DESCRIPTION";
  static const std::map<int, std::string> descriptions = {
    {StatusCode::INITIAL,
     "The transaction has just been intialized. It has not been executed yet "
     "and clients can still change it and add sub-transactions."},
    {StatusCode::GO_ON,
     "The transaction has a transaction_type yet and clients may add matching "
     "sub-transaction or execute it right-away."},
    {StatusCode::READY,
     "The transaction is ready for execution and cannot be changed anymore."},
    {StatusCode::EXECUTING, "The transaction is currently being executed."},
    {StatusCode::SUCCESS, "The action was successful"},
    {StatusCode::CONNECTION_ERROR,
     "The attempt to execute this transaction was not successful because the "
     "connection to the server could not be established."},
    {StatusCode::AUTHENTICATION_ERROR,
     "The attempt to execute this transaction has not been executed at all "
     "because the authentication did not succeed."},
    {StatusCode::GENERIC_RPC_ERROR,
     "The attempt to execute this transaction was not successful because an "
     "error occured in the transport or RPC protocol layer."},
    {StatusCode::GENERIC_ERROR,
     "An error occured. Please review the logs for more information."},
    {StatusCode::GENERIC_TRANSACTION_ERROR,
     "The transaction terminated unsuccessfully with transaction errors."},
    {StatusCode::CONFIGURATION_ERROR,
     "An error occurred during the configuration of the ConfigurationManager."},
    {StatusCode::UNKNOWN_CONNECTION_ERROR,
     "The ConnectionManager does not know any connection of this name."},
    {StatusCode::TRANSACTION_STATUS_ERROR,
     "The Transaction is in a wrong state for the attempted action."},
    {StatusCode::TRANSACTION_TYPE_ERROR,
     "The Transaction has a transaction type which does not allow the "
     "attempted action."},
    {StatusCode::ORIGINAL_ENTITY_MISSING_ID,
     "The attempt to update this entity failed because this entity does not "
     "have "
     "an id. This is the case when you did not retrieve it before applying any "
     "changes and instantiated the Entity class explicitely."},
    {StatusCode::NOT_A_FILE_ENTITY, "You can only add files to file entities."},
    {StatusCode::PATH_IS_A_DIRECTORY, "The given path is a directory."},
    {StatusCode::FILE_DOES_NOT_EXIST_LOCALLY,
     "The file does not not exist in the local file system."},
    {StatusCode::FILE_DOWNLOAD_ERROR,
     "The transaction failed during the download of the files"},
    {StatusCode::FILE_UPLOAD_ERROR,
     "The transaction failed during the upload of the files"},
    {StatusCode::UNSUPPORTED_FEATURE,
     "This feature is not available in the this client implementation."},
    {StatusCode::EXTERN_C_ASSIGNMENT_ERROR,
     "You tried to assign a new object to the wrapped void pointer. You have "
     "to delete the old pointee first."},
    {StatusCode::OTHER_CLIENT_ERROR,
     "This is code is reserved to errors raised by other clients wrapping the "
     "C++ client (or its Extern C interface).  This should never occur when "
     "working with the C++ code itself."}};
  try {
    return descriptions.at(code);
  } catch (const std::out_of_range &exc) {
    return MISSING_DESCRIPTION;
  }
}

} // namespace caosdb

namespace caosdb::transaction {
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::FileTransmissionService;
using caosdb::entity::v1alpha1::MultiTransactionRequest;
using caosdb::entity::v1alpha1::MultiTransactionResponse;
using TransactionResponseCase =
  caosdb::entity::v1alpha1::TransactionResponse::TransactionResponseCase;
using RetrieveResponseCase =
  caosdb::entity::v1alpha1::RetrieveResponse::RetrieveResponseCase;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using google::protobuf::Arena;
using NextStatus = grpc::CompletionQueue::NextStatus;
using RegistrationStatus = caosdb::entity::v1alpha1::RegistrationStatus;

ResultSet::iterator::iterator(const ResultSet *result_set_param, int index)
  : current_index(index), result_set(result_set_param) {}

auto ResultSet::iterator::operator*() const -> const Entity & {
  return this->result_set->at(current_index);
}

auto ResultSet::iterator::operator++() -> ResultSet::iterator & {
  current_index++;
  return *this;
}

auto ResultSet::iterator::operator++(int) -> ResultSet::iterator {
  iterator tmp(*this);
  operator++();
  return tmp;
}

auto ResultSet::iterator::operator!=(const iterator &rhs) const -> bool {
  return this->current_index != rhs.current_index;
}

auto ResultSet::begin() const -> ResultSet::iterator {
  return ResultSet::iterator(this, 0);
}

auto ResultSet::end() const -> ResultSet::iterator {
  return ResultSet::iterator(this, size());
}

MultiResultSet::MultiResultSet(std::vector<std::unique_ptr<Entity>> result_set)
  : AbstractMultiResultSet(std::move(result_set)) {}

Transaction::Transaction(
  std::shared_ptr<EntityTransactionService::Stub> entity_service,
  std::shared_ptr<FileTransmissionService::Stub> file_service)
  : request(Arena::CreateMessage<MultiTransactionRequest>(GetArena())),
    response(Arena::CreateMessage<MultiTransactionResponse>(GetArena())) {
  this->entity_service = std::move(entity_service);
  this->file_service = std::move(file_service);
  this->query_count = -1;
}

auto Transaction::RetrieveById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->set_id(id);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::RetrieveAndDownloadFilesById(
  const std::string &id, const std::string &local_path) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *retrieve_request =
    this->request->add_requests()->mutable_retrieve_request();
  retrieve_request->set_id(id);
  retrieve_request->set_register_file_download(true);

  download_files[id].local_path = local_path;

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::Query(const std::string &query) noexcept -> StatusCode {
  ASSERT_CAN_ADD_QUERY

  // currently, the server can only process one query at a time (but mixed with
  // other retrievals).
  this->has_query = true;
  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->mutable_query()->set_query(query);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::DeleteById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_DELETION

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_delete_request()->set_id(id);

  this->status = TransactionStatus::GO_ON();
  return this->status.GetCode();
}

auto Transaction::InsertEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_INSERTION

  auto *entity_request = this->request->add_requests()
                           ->mutable_insert_request()
                           ->mutable_entity_request();
  auto *proto_entity = entity_request->mutable_entity();

  // copy the original entity for the transaction
  entity->CopyTo(proto_entity);
  if (entity->HasFile()) {
    auto *file_transmission_id = entity_request->mutable_upload_id();
    entity->SetFileTransmissionId(file_transmission_id);
    upload_files.push_back(entity->GetFileDescriptor());
  }
  this->status = TransactionStatus::READY();
  return this->status.GetCode();
}

auto Transaction::UpdateEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_UPDATE

  auto *entity_request = this->request->add_requests()
                           ->mutable_update_request()
                           ->mutable_entity_request();
  auto *proto_entity = entity_request->mutable_entity();

  entity->CopyTo(proto_entity);
  if (entity->HasFile()) {
    auto *file_transmission_id = entity_request->mutable_upload_id();
    entity->SetFileTransmissionId(file_transmission_id);
    upload_files.push_back(entity->GetFileDescriptor());
  }
  this->status = TransactionStatus::READY();
  return this->status.GetCode();
}

auto Transaction::Execute() -> TransactionStatus {
  auto status_code = ExecuteAsynchronously();
  TransactionStatus::ThrowExceptionIfError(
    status_code, caosdb::get_status_description(status_code));
  auto status = WaitForIt();
  status.ThrowExceptionIfError();
  return status;
}

// TODO(tf) This has apparently a cognitive complexity of 39>25 (threshold).
auto Transaction::ExecuteAsynchronously() noexcept -> StatusCode { // NOLINT
  if (!IsStatus(TransactionStatus::READY()) &&
      !IsStatus(TransactionStatus::GO_ON())) {
    return StatusCode::TRANSACTION_STATUS_ERROR;
  }
  switch (this->transaction_type) {
  case MIXED_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed write transactions (containing insertions, deletions, and updates "
      "in one transaction).")
  case MIXED_READ_AND_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed read and write transactions (containing retrievals, insertions, "
      "deletions, and updates in one transaction).")
  default:
    break;
  }
  this->status = TransactionStatus::EXECUTING();

  // upload files first
  if (!upload_files.empty()) {
    CAOSDB_LOG_INFO(logger_name)
      << "Number of files to be uploaded: " << upload_files.size();

    auto *registration_request =
      Arena::CreateMessage<RegisterFileUploadRequest>(GetArena());
    auto *registration_response =
      Arena::CreateMessage<RegisterFileUploadResponse>(GetArena());

    handler_ = std::make_unique<RegisterFileUploadHandler>(
      &handler_, file_service.get(), &completion_queue, registration_request,
      registration_response);
    this->status = ProcessCalls();

    if (registration_response->status() !=
        RegistrationStatus::REGISTRATION_STATUS_ACCEPTED) {
      this->status = TransactionStatus::FILE_UPLOAD_ERROR();
      return StatusCode::EXECUTING;
    }

    for (auto &file_descriptor : upload_files) {
      file_descriptor.file_transmission_id->set_registration_id(
        registration_response->registration_id());
      CAOSDB_LOG_INFO(logger_name)
        << "Uploading " << file_descriptor.local_path;
      handler_ = std::make_unique<UploadRequestHandler>(
        &handler_, file_service.get(), &completion_queue, file_descriptor);
      this->status = ProcessCalls();
      if (this->status.GetCode() != StatusCode::EXECUTING) {
        return StatusCode::EXECUTING;
      }
    }
  }

  CAOSDB_LOG_DEBUG(logger_name) << "RPC Request: " << RequestToString();
  handler_ = std::make_unique<EntityTransactionHandler>(
    &handler_, entity_service.get(), &completion_queue, request, response);
  this->status = ProcessCalls();
  if (this->status.GetCode() != StatusCode::EXECUTING) {
    return StatusCode::EXECUTING;
  }

  // file download afterwards
  if (status.GetCode() == StatusCode::EXECUTING && !download_files.empty()) {
    // run over all retrieved entities and get the download_id
    for (auto sub_response : *(response->mutable_responses())) {
      if (sub_response.transaction_response_case() ==
          TransactionResponseCase::kRetrieveResponse) {
        if (sub_response.retrieve_response()
              .entity_response()
              .has_download_id()) {
          auto *entity_response =
            sub_response.mutable_retrieve_response()->mutable_entity_response();
          auto entity_id = entity_response->entity().id();
          download_files[entity_id].file_transmission_id =
            entity_response->release_download_id();
          // TODO(tf) handle error
        }
      }
    }

    for (const auto &item : download_files) {
      auto file_descriptor(item.second);
      CAOSDB_LOG_INFO(logger_name)
        << "Downloading " << file_descriptor.local_path;

      handler_ = std::make_unique<DownloadRequestHandler>(
        &handler_, file_service.get(), &completion_queue, file_descriptor);
      this->status = ProcessCalls();
      if (this->status.GetCode() != StatusCode::EXECUTING) {
        return StatusCode::EXECUTING;
      }
    }
  }
  return StatusCode::EXECUTING;
}

// TODO(tf) This has apparently a cognitive complexity of 36>25 (threshold).
auto Transaction::WaitForIt() const noexcept -> TransactionStatus { // NOLINT
  if (this->status.GetCode() != StatusCode::EXECUTING) {
    return this->status;
  }
  this->status = TransactionStatus::SUCCESS();
  bool set_error = false;
  auto *responses = this->response->mutable_responses();
  std::vector<std::unique_ptr<Entity>> entities;
  for (auto &sub_response : *responses) {
    std::unique_ptr<Entity> result;
    switch (sub_response.transaction_response_case()) {

    case TransactionResponseCase::kRetrieveResponse: {
      auto *retrieve_response = sub_response.mutable_retrieve_response();

      switch (retrieve_response->retrieve_response_case()) {
      case RetrieveResponseCase::kEntityResponse: {
        auto *retrieve_entity_response =
          retrieve_response->release_entity_response();
        result = std::make_unique<Entity>(retrieve_entity_response);
      } break;
      case RetrieveResponseCase::kSelectResult: {
        CAOSDB_LOG_ERROR(logger_name) << "Results of a SELECT query cannot be "
                                         "processed by this client yet.";
        // TODO(tf) Select queries
      } break;
      case RetrieveResponseCase::kCountResult: {
        this->query_count = retrieve_response->count_result();
      } break;
      case RetrieveResponseCase::kFindResult: {
        std::unique_ptr<Entity> find_result;
        for (auto &entity_response :
             *retrieve_response->mutable_find_result()->mutable_result_set()) {
          find_result = std::make_unique<Entity>(&entity_response);
          if (find_result->HasErrors()) {
            set_error = true;
          }
          entities.push_back(std::move(find_result));
        }
      } break;
      default:
        CAOSDB_LOG_FATAL(logger_name) << "Received invalid QueryResponseCase.";
        break;
      }

      break; // break TransactionResponseCase::kRetrieveResponse
    }

    case TransactionResponseCase::kInsertResponse: {
      auto *inserted_id_response =
        sub_response.mutable_insert_response()->release_id_response();
      result = std::make_unique<Entity>(inserted_id_response);
      break;
    }
    case TransactionResponseCase::kDeleteResponse: {
      auto *deleted_id_response =
        sub_response.mutable_delete_response()->release_id_response();
      result = std::make_unique<Entity>(deleted_id_response);
      break;
    }
    case TransactionResponseCase::kUpdateResponse: {
      auto *updated_id_response =
        sub_response.mutable_update_response()->release_id_response();
      result = std::make_unique<Entity>(updated_id_response);
      break;
    }
    default:
      CAOSDB_LOG_FATAL(logger_name)
        << "Received invalid TransactionResponseCase.";
      break;
    }
    if (result != nullptr) {
      if (result->HasErrors()) {
        set_error = true;
      }
      entities.push_back(std::move(result));
    }
  }

  // copy local path of downloaded files into the entities file descriptor
  for (auto &entity : entities) {
    auto id = entity->GetId();
    if (!id.empty() && download_files.count(id) == 1) {
      const auto &local_path = download_files.at(id).local_path;
      entity->SetLocalPath(local_path);
    }
  }
  this->result_set = std::make_unique<MultiResultSet>(std::move(entities));

  if (set_error) {
    this->status = TransactionStatus::TRANSACTION_ERROR(
      "The request terminated with errors.");
  }

  return this->status;
}

auto Transaction::ProcessCalls() -> TransactionStatus {
  gpr_timespec deadline;
  deadline.tv_sec = 1;
  deadline.tv_nsec = 0;
  deadline.clock_type = gpr_clock_type::GPR_TIMESPAN;

  TransactionStatus result = TransactionStatus::EXECUTING();
  handler_->Start();
  void *tag = nullptr;
  bool ok = false;
  while (true) {
    switch (completion_queue.AsyncNext(&tag, &ok, deadline)) {
    case NextStatus::GOT_EVENT: {
      if (tag != nullptr) {
        auto res = handler_->OnNext(ok);
        if (!res) {
          // The handler has finished it's work
          result = handler_->GetStatus();
          handler_.reset();
          return result;
        }
      } else {
        std::string description("Invalid tag delivered by notification queue.");
        CAOSDB_LOG_ERROR(logger_name) << description;
        handler_.reset();
        return TransactionStatus::RPC_ERROR(description);
      }
    } break;
    case NextStatus::SHUTDOWN: {
      CAOSDB_LOG_ERROR(logger_name)
        << "Notification queue has been shut down unexpectedly.";
      result = handler_->GetStatus();
      handler_.reset();
      return result;
    } break;
    case NextStatus::TIMEOUT: {
      CAOSDB_LOG_DEBUG(logger_name) << "Timeout, waiting...";
    } break;
    default:
      CAOSDB_LOG_FATAL(logger_name)
        << "Got an invalid NextStatus from CompletionQueue.";
      result = handler_->GetStatus();
      handler_.reset();
      return result;
    }
  }
  result = handler_->GetStatus();
  handler_.reset();
  return result;
}

Transaction::~Transaction() {
  this->Cancel();

  completion_queue.Shutdown();

  // drain the queue
  void *ignoredTag = nullptr;
  bool ok = false;
  while (completion_queue.Next(&ignoredTag, &ok)) {
    ;
  }
}

void Transaction::Cancel() {
  // TODO(tf) State Canceled
  if (handler_ != nullptr) {
    handler_->Cancel();
  }
}

} // namespace caosdb::transaction
